#include "xpulse.h"
#include "extern.h"
#include <X11/keysym.h>
#include <X11/keysymdef.h>

extern void buildGauge();

/*
 *  Determine system load in range between [0 .. MAXLOAD]
 */

void getLoad(Load *load, Load *last)
{
  Load current;
  unsigned long denom;
  float avg;
  char *str;

  if((lseek(loadfd, (off_t)0, SEEK_SET) != 0) ||
    ((read(loadfd, filebuffer, 80)) <= 0)) {
    (void)perror("reading /proc/stat");
    cleanUp();
    exit(EXIT_FAILURE);
  }
  str = filebuffer;
  while(!isdigit((int)*str))
    str++;
  if(sscanf(str, "%d %d %d %d\n", &current.user, &current.nice,
      &current.system, &current.idle) != 4) {
    (void)fprintf(stderr, "%s: /proc/stat: bad file format\n");
    cleanUp();
    exit(EXIT_FAILURE);
  }
  current.total = current.user + current.nice + current.system;

  load->total = current.total - last->total;
  last->total = current.total;
  load->user = current.user - last->user;
  last->user = current.user;
  load->system = current.system - last->system;
  last->system = current.system;
  load->nice = current.nice - last->nice;
  last->nice = current.nice;
  load->idle = current.idle - last->idle;
  last->idle = current.idle;

  if((modeswitch & MODE_DEBUG)
      && (load->total <= 999) && (load->user <= 999)
      && (load->system <= 999) && (load->nice <= 999))
    (void)fprintf(stderr,
      "\rtotal:%3d, user:%3d, system:%3d, nice:%3d",
      load->total, load->user, load->system, load->nice);
  
  if(twinmode == TWIN_AVERAGE) {
    if((lseek(avgfd, (off_t)0, SEEK_SET) != 0) ||
      ((read(avgfd, filebuffer, 80)) <= 0)) {
      (void)perror("reading /proc/loadavg");
      cleanUp();
      exit(EXIT_FAILURE);
    }
    str = filebuffer;
    while(!isdigit((int)*str))
      str++;
    if(sscanf(str, "%f", &avg) != 1) {
      (void)fprintf(stderr, "%s: /proc/loadavg: bad file format\n");
      cleanUp();
      exit(EXIT_FAILURE);
    }
    /*
    if(modeswitch & MODE_DEBUG)
      (void)fprintf(stderr, ", average: %.2f", avg);
    */
    /*
    load->average = avg * 10;
    */
    load->average = avg * 11;
  }
  else
    load->average = 0;

  denom = load->total + load->idle;
  load->total = (load->total * multiply) / denom;
  load->user = (load->user * multiply) / denom;
  load->system = (load->system * multiply) / denom;
  load->nice = (load->nice * multiply) / denom;
  load->idle = (load->idle * multiply) / denom;

  if(load->total > MAXLOAD)
    load->total = MAXLOAD;
  if(load->user > MAXLOAD)
    load->user = MAXLOAD;
  if(load->system > MAXLOAD)
    load->system = MAXLOAD;
  if(load->nice > MAXLOAD)
    load->nice = MAXLOAD;
  if(load->idle > MAXLOAD)
    load->idle = MAXLOAD;
  if(load->average > MAXLOAD)
    load->average = MAXLOAD;
}

/*
 *  Draw pre-painted CPU-pixmap
 */

void drawImage(int x, int y)
{
  if(modeswitch & MODE_CLASSIC)
    XSetForeground(display, maingc, greypixel);
  else
    XSetForeground(display, maingc, bgpixel);
  XFillRectangle(display, mainwin, maingc,
    x, y, (x + IMAGESIZE + 4), (y + IMAGESIZE + 4));
  XCopyArea(display, pixmap, mainwin, maingc,
    0, 0, IMAGESIZE, IMAGESIZE, OFFSET, OFFSET);
}

/*
 *  Draw 3D-border around meter area
 */

void drawGaugeBorder(int x, int y)
{
  int a;
  int b;

  XSetForeground(display, maingc, greypixel);
  XFillRectangle(display, mainwin, maingc,
    x, y, (winwidth - x), (winheight - y));
  XSetForeground(display, maingc, bgpixel);
  a  = (OFFSET + IMAGESIZE + 6 + 2);
  b = ((winheight / 2) - (GAUGEHEIGHT / 2) - 2);
  XFillRectangle(display, mainwin, maingc, a, b,
    (2 + (MAXLOAD * 10)), (GAUGEHEIGHT + 4 + 1));
  XSetForeground(display, maingc, greyshade[2]);
  XFillRectangle(display, mainwin, maingc, (a - 2), (b - 2),
    (2 + (MAXLOAD * 10) + 3), 2);
  XFillRectangle(display, mainwin, maingc, (a - 2), b,
    2, (GAUGEHEIGHT + 5));
  XSetForeground(display, maingc, whitepixel);
  XFillRectangle(display, mainwin, maingc, (a - 1),
    (b + GAUGEHEIGHT + 4), (2 + (MAXLOAD * 10) + 3), 2);
  XFillRectangle(display, mainwin, maingc,
    (a + 2 + (MAXLOAD * 10)), (b - 1), 2, (GAUGEHEIGHT + 7));
}

/*
 *  Draw level meter
 */

void drawGauge(int x, int y, int load, unsigned long table[])
{
  int i;

  for(i = 0; i < load; i++) {
    XSetForeground(display, gaugegc, table[i]);
    XFillRectangle(display, mainwin, gaugegc,
      (i * 10 + x), y, 8, GAUGEHEIGHT);
  }
  if(load < MAXLOAD) {
    if(modeswitch & MODE_METER) {
      XSetForeground(display, gaugegc, greyshade[0]);
      for(1; i < MAXLOAD; i++)
        XFillRectangle(display, mainwin, gaugegc,
          (i * 10 + x), y, 8, GAUGEHEIGHT);
    }
    else
      XClearArea(display, mainwin, (i * 10 + x), y,
        ((MAXLOAD - i) * 10), GAUGEHEIGHT, False);
  }
}

/*
 *  Interrupt-handler
 */

void timerLoop(int sig)
{
  Load load;
  static Load loadsave;

  if(modeswitch & MODE_NOP)
    return;
  getLoad(&load, &loadsave);
  if(modeswitch & MODE_CLASSIC)
    drawGauge((OFFSET + IMAGESIZE + 6 + 4),
      ((winheight / 2) - (GAUGEHEIGHT / 2)),
	load.total, hicolor);
  else if((modeswitch & MODE_TWIN) && (twinmode == TWIN_USERSYS)) {
    drawGauge((OFFSET + IMAGESIZE + 6),
      ((winheight / 2) - GAUGEHEIGHT - 2),
	load.user, hicolor);
    drawGauge((OFFSET + IMAGESIZE + 6),
      ((winheight / 2) + 2),
	load.system, locolor);
  }
  else if((modeswitch & MODE_TWIN) && (twinmode == TWIN_AVERAGE)) {
    drawGauge((OFFSET + IMAGESIZE + 6),
      ((winheight / 2) - GAUGEHEIGHT - 2),
	load.total, hicolor);
    drawGauge((OFFSET + IMAGESIZE + 6),
      ((winheight / 2) + 2),
	load.average, locolor);
  }
  else
    drawGauge((OFFSET + IMAGESIZE + 6),
      ((winheight / 2) - (GAUGEHEIGHT / 2)),
	load.total, hicolor);
  XFlushGC(display, gaugegc);
  XFlush(display);
  return;
}

/*
 *  Main Xlib-event-evaluation loop
 */

void mainLoop()
{
  XEvent event;
  int loop;

  loop = 1;

  while(loop) {
    XNextEvent(display, &event);
    switch(event.type) {
    case Expose:
      if(event.xexpose.count == 0) {
	if(modeswitch & MODE_CLASSIC)
	  drawGaugeBorder((OFFSET + IMAGESIZE + OFFSET), 0);
	drawImage(0, 0);
      }
      break;
    case MappingNotify:
      XRefreshKeyboardMapping(&event.xmapping);
      break;
    case KeyPress:
	if(XKeycodeToKeysym(display, event.xkey.keycode, 0) ==
	    XK_Escape) {
	  if(modeswitch & MODE_DEBUG)
	    (void)fprintf(stderr, "\n");
	  loop = 0;
	}
	break;
    case ButtonPress:
      if((event.xbutton.button == Button1)
        && (event.xbutton.x > OFFSET)
	&& (event.xbutton.x < (IMAGESIZE + OFFSET))
        && (event.xbutton.y > OFFSET)
	&& (event.xbutton.y < (IMAGESIZE + OFFSET)))
        openInfoWindow(event.xbutton.x_root, event.xbutton.y_root);
      else if((event.xbutton.x > (OFFSET + IMAGESIZE + OFFSET))
	  && (modeswitch & MODE_TWIN)
	  && (modeswitch & MODE_SWITCH)
	  && (event.xbutton.button == Button1)) {
        if(twinmode == TWIN_USERSYS)
	  twinmode = TWIN_AVERAGE;
        else if(twinmode == TWIN_AVERAGE)
	  twinmode = TWIN_USERSYS;
        buildGauge();
      }
      break;
    case ClientMessage:
      if((event.xclient.message_type == wmprot) &&
	  (event.xclient.data.l[0] == wmdel)) {
	if(modeswitch & MODE_DEBUG)
	  (void)fprintf(stderr, "\n");
	loop = 0;
      }
      break;
    }
  }
}
